#! /usr/bin/env python
"""Generate index page for the rendered prefix trees."""
import json
import os
import pathlib
import sys
from typing import Collection, Dict, List, Union

import mapology.db as db
import mapology.template_loader as template
from mapology import BASE_URL, ENCODING, FOOTER_HTML, FS_PREFIX_PATH, LIB_PATH, PATH_NAV, log

FeatureDict = Dict[str, Collection[str]]
PHeaderDict = Dict[str, Collection[str]]
PFeatureDict = Dict[str, Collection[str]]

HTML_TEMPLATE = os.getenv('GEO_INDEX_HTML_TEMPLATE', '')
HTML_TEMPLATE_IS_EXTERNAL = bool(HTML_TEMPLATE)
if not HTML_TEMPLATE:
    HTML_TEMPLATE = 'prefix_index_template.html'

AIRP = 'airport'
RUNW = 'runways'
FREQ = 'frequencies'
LOCA = 'localizers'
GLID = 'glideslopes'

CC_HINT = 'CC_HINT'
City = 'City'
CITY = City.upper()
ICAO = 'ICAO'
IC_PREFIX = 'IC_PREFIX'
ITEM = 'ITEM'
KIND = 'KIND'
PATH = 'PATH'
BASE_URL_TARGET = 'BASE_URL'
ANCHOR = 'ANCHOR'
TEXT = 'TEXT'
URL = 'URL'
ZOOM = 'ZOOM'
DEFAULT_ZOOM = 1
FOOTER_HTML_KEY = 'FOOTER_HTML'
LIB_PATH_KEY = 'LIB_PATH'

icao = 'icao_lower'
ic_prefix_token = 'ic_prefix_lower'  # nosec B105
LAT_LON = 'LAT_LON'
cc_page = 'cc_page'
Cc_page = 'Cc_page'

ATTRIBUTION = f'{KIND} {ITEM} of '

# GOOGLE_MAPS_URL = f'https://www.google.com/maps/search/?api=1&query={{lat}}%2c{{lon}}'  # Map + pin Documented
GOOGLE_MAPS_URL = 'https://maps.google.com/maps?t=k&q=loc:{lat}+{lon}'  # Sat + pin Undocumented


def main(argv: Union[List[str], None] = None) -> int:
    """Drive the derivation."""
    argv = sys.argv[1:] if argv is None else argv
    if argv:
        print('usage: mapology index')
        return 2

    store_index = db.load_index('store')
    table_index = db.load_index('table')
    apt_search_index = db.load_index('apt_search')

    prefixes = sorted(table_index.keys())
    row_slot_set = set(prefix[0] for prefix in prefixes)
    col_slot_set = set(prefix[1] for prefix in prefixes)
    columns = sorted(col_slot_set)
    rows = sorted(row_slot_set)

    table = {f'{k}{j}': '<td>&nbsp;</td>' for k in rows for j in columns}
    log.debug('I\\C |' + ' | '.join(columns))
    for row in rows:
        log.debug(row)

    regions, total_airports = 0, 0
    for prefix in store_index:
        with open(store_index[prefix], 'rt', encoding=ENCODING) as handle:
            prefix_store = json.load(handle)

        with open(table_index[prefix], 'rt', encoding=ENCODING) as handle:
            table_store = json.load(handle)

        regions += 1
        count = len(prefix_store['features'])
        total_airports += count
        region_name = table_store['name']
        table[prefix] = f'<td><a href="{prefix}/" class="nd" title="{region_name}">{count}</a></td>'

    tr_pad = 12
    sp = ' '
    th_pad = tr_pad + 2
    data_cols = [f'{sp * tr_pad}<tr>', f'{sp * th_pad}<th>I\\C</th>']
    for col in columns:
        data_cols.append(f'{sp * th_pad}<th>{col}</th>')
    data_cols.append(f'{sp * tr_pad}</tr>')

    data_rows = []
    for k in rows:
        row = [f'<tr><th>{k}</th>']
        for j in columns:
            row.append(table[f'{k}{j}'])
        row.append('\n')
        data_rows.append(''.join(row))

    html_dict = {
        ANCHOR: 'prefix/',
        LIB_PATH_KEY: LIB_PATH,
        PATH: PATH_NAV,
        LAT_LON: '0, 0',
        ZOOM: str(DEFAULT_ZOOM),
        BASE_URL_TARGET: BASE_URL,
        'NUMBER_REGIONS': str(regions),
        'TOTAL_AIRPORTS': str(total_airports),
        FOOTER_HTML_KEY: FOOTER_HTML,
        'DATA_COLS': '\n'.join(data_cols),
        'DATA_ROWS': ''.join(data_rows),
    }
    html_page = template.load_html(HTML_TEMPLATE, HTML_TEMPLATE_IS_EXTERNAL)
    for key, replacement in html_dict.items():
        html_page = html_page.replace(key, replacement)

    html_path = pathlib.Path(FS_PREFIX_PATH, 'index.html')
    with open(html_path, 'wt', encoding=ENCODING) as html_handle:
        html_handle.write(html_page)

    search_index = []
    for prefix in apt_search_index:
        with open(apt_search_index[prefix], 'rt', encoding=ENCODING) as handle:
            search_index.append(json.load(handle))

    search_index_path = pathlib.Path('apt-search-index.json')
    with open(search_index_path, 'wt', encoding=ENCODING) as handle:
        json.dump(search_index, handle, indent=2)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



"""Loader function for templates."""
import importlib.resources
import pathlib
from typing import List, Union

ENCODING = 'utf-8'
RESOURCES = (
    'airport_page_template.html',
    'country_index_template.html',
    'country_page_template.html',
    'ndb_index_template.html',
    'prefix_index_template.html',
    'prefix_page_template.html',
    'search_index_template.html',
)


def load_html(resource: str, is_complete_path: bool = False) -> str:
    """Load the template either from the package resources or an external path."""
    if is_complete_path:
        with open(resource, 'rt', encoding=ENCODING) as handle:
            return handle.read()
    else:
        with importlib.resources.path(__package__, resource) as html_template_path:
            with open(html_template_path, 'rt', encoding=ENCODING) as handle:
                return handle.read()


def eject(argv: Union[List[str], None] = None) -> int:
    """Eject the templates into the folder given (default EJECTED) and create the folder if it does not exist."""
    argv = argv if argv else ['']
    into = argv[0]
    if not into.strip():
        into = 'EJECTED'
    into_path = pathlib.Path(into)
    into_path.mkdir(parents=True, exist_ok=True)
    for resource in RESOURCES:
        write_to = into_path / resource
        with importlib.resources.path(__package__, resource) as html_template_path:
            with open(html_template_path, 'rt', encoding=ENCODING) as handle:
                with open(write_to, 'wt', encoding=ENCODING) as target:
                    target.write(handle.read())

    return 0



"""Database for the complete tree."""
import copy
import json
import pathlib
from typing import Collection, Dict, Mapping, no_type_check

from mapology import ENCODING, FS_DB_ROOT_PATH

FS_DB_STORE_PART = 'prefix-store'
FS_DB_TABLE_PART = 'prefix-table'
FS_DB_HULLS_PART = 'prefix-hulls'
FS_DB_APT_SEARCH_PART = 'prefix-apt-search'
FS_DB_REGION_COUNTRY_PART = 'prefix-country-region'

DB_ROOT = pathlib.Path(FS_DB_ROOT_PATH)
DB_FOLDER_PATHS = {
    'hulls': DB_ROOT / FS_DB_HULLS_PART,
    'store': DB_ROOT / FS_DB_STORE_PART,
    'table': DB_ROOT / FS_DB_TABLE_PART,
    'apt_search': DB_ROOT / FS_DB_APT_SEARCH_PART,
    'region_country': DB_ROOT / FS_DB_REGION_COUNTRY_PART,
}

DB_INDEX_PATHS = {
    'hulls': DB_ROOT / f'{FS_DB_HULLS_PART}.json',
    'store': DB_ROOT / f'{FS_DB_STORE_PART}.json',
    'table': DB_ROOT / f'{FS_DB_TABLE_PART}.json',
    'apt_search': DB_ROOT / f'{FS_DB_APT_SEARCH_PART}.json',
    'region_country': DB_ROOT / f'{FS_DB_REGION_COUNTRY_PART}.json',
}
PHeaderDict = Dict[str, Collection[str]]
CC_HINT = 'CC_HINT'
IC_PREFIX = 'IC_PREFIX'
GEO_JSON_PREFIX_HEADER: PHeaderDict = {
    'type': 'FeatureCollection',
    'id': f'{IC_PREFIX}',
    'name': f'Region - {IC_PREFIX} ({CC_HINT})',
    'crs': {
        'type': 'name',
        'properties': {
            'name': 'urn:ogc:def:crs:OGC:1.3:CRS84',
        },
    },
    'features': [],
}
JSON_PREFIX_TABLE_HEADER = {
    'type': 'x-prefix-table',
    'id': f'{IC_PREFIX}',
    'name': f'Region - {IC_PREFIX} ({CC_HINT})',
    'crs': {
        'type': 'name',
        'properties': {
            'name': 'urn:ogc:def:crs:OGC:1.3:CRS84',
        },
    },
    'airports': [],
}

JSON_PREFIX_APT_SEARCH_HEADER = []  # type: ignore
JSON_PREFIX_COUNTRY_HEADER = {}  # type: ignore


def ensure_fs_tree() -> None:
    """Ensure the DB folder tree exists."""
    for db in DB_FOLDER_PATHS.values():
        db.mkdir(parents=True, exist_ok=True)

    for index in DB_INDEX_PATHS.values():
        if not index.exists():
            with open(index, 'wt', encoding=ENCODING) as handle:
                json.dump({}, handle, indent=2)


@no_type_check
def load_index(kind: str) -> Mapping[str, str]:
    """DRY."""
    with open(DB_INDEX_PATHS[kind], 'rt', encoding=ENCODING) as handle:
        return json.load(handle)


def dump_index(kind: str, data: Mapping[str, str]) -> None:
    """DRY."""
    with open(DB_INDEX_PATHS[kind], 'wt', encoding=ENCODING) as handle:
        json.dump(data, handle, indent=2)


def hull_path(some_prefix: str) -> str:
    """DRY."""
    return str(DB_FOLDER_PATHS['hulls'] / f'{some_prefix}.json')


def apt_search_path(some_prefix: str) -> str:
    """DRY."""
    return str(DB_FOLDER_PATHS['apt_search'] / f'{some_prefix}.json')


def region_country_path(some_prefix: str) -> str:
    """DRY."""
    return str(DB_FOLDER_PATHS['region_country'] / f'{some_prefix}.json')


def add_prefix(icp: str, cc: str) -> PHeaderDict:
    """DRY."""
    geojson = copy.deepcopy(GEO_JSON_PREFIX_HEADER)
    geojson['name'] = geojson['name'].replace(IC_PREFIX, icp).replace(CC_HINT, cc)  # type: ignore
    geojson['id'] = geojson['id'].replace(IC_PREFIX, icp)  # type: ignore
    return geojson


def add_table_prefix(icp: str, cc: str) -> PHeaderDict:
    """DRY."""
    table = copy.deepcopy(JSON_PREFIX_TABLE_HEADER)
    table['name'] = table['name'].replace(IC_PREFIX, icp).replace(CC_HINT, cc)  # type: ignore
    table['id'] = table['id'].replace(IC_PREFIX, icp)  # type: ignore
    return table


@no_type_check
def update_aspect(index_db: Mapping[str, object], a_prefix: str, a_cc: str, kind: str) -> Mapping[str, object]:
    """Process the kinds' index and ensure prefix aspect db is present"""
    if a_prefix not in index_db:
        index_db[a_prefix] = str(DB_FOLDER_PATHS[kind] / f'{a_prefix}.json')  # noqa
        # Create initial kinds' store data entry for ICAO prefix
        factory = add_prefix if kind == 'store' else add_table_prefix
        with open(index_db[a_prefix], 'wt', encoding=ENCODING) as handle:  # noqa
            json.dump(factory(a_prefix, a_cc), handle)
    with open(index_db[a_prefix], 'rt', encoding=ENCODING) as handle:  # noqa
        return json.load(handle)



"""Simple python filters generating leaflet driven apps."""
import logging
import os
import pathlib
from typing import List, no_type_check

APP_ALIAS = str(pathlib.Path(__file__).parent.name)
APP_ENV = APP_ALIAS.upper()
APP_NAME = locals()['__doc__']
DEBUG = bool(os.getenv(f'{APP_ENV}_DEBUG', ''))
ENCODING = 'utf-8'
ENCODING_ERRORS_POLICY = 'ignore'
DEFAULT_CONFIG_NAME = f'.{APP_ALIAS}.json'

ENCODING = 'utf-8'

COUNTRY_PAGE = os.getenv('GEO_COUNTRY_PAGE', '')
PATH_NAV = os.getenv('GEO_PATH_NAV', '')
LIB_PATH = os.getenv('GEO_LIB_PATH', '/ndb/_/leaflet')
BASE_URL = os.getenv('BASE_URL', 'http://localhost:8080')
FOOTER_HTML = os.getenv('GEO_FOOTER_HTML', ' ')
FS_PREFIX_PATH = os.getenv('GEO_PREFIX_PATH', 'prefix')
FS_DB_ROOT_PATH = os.getenv('GEO_DB_ROOT_PATH', 'db')

log = logging.getLogger()  # Temporary refactoring: module level logger
LOG_FOLDER = pathlib.Path('logs')
LOG_FILE = f'{APP_ALIAS}.log'
LOG_PATH = pathlib.Path(LOG_FOLDER, LOG_FILE) if LOG_FOLDER.is_dir() else pathlib.Path(LOG_FILE)
LOG_LEVEL = logging.INFO

# [[[fill git_describe()]]]
__version__ = '2023.10.21+parent.173bf20d'
# [[[end]]] (checksum: 3586d386a70db02e75f620206a4dc824)
__version_info__ = tuple(
    e if '-' not in e else e.split('-')[0] for part in __version__.split('+') for e in part.split('.') if e != 'parent'
)
__all__: List[str] = []


@no_type_check
def init_logger(name=None, level=None):
    """Initialize module level logger"""
    global log  # pylint: disable=global-statement

    log_format = {
        'format': '%(asctime)s.%(msecs)03d %(levelname)s [%(name)s]: %(message)s',
        'datefmt': '%Y-%m-%dT%H:%M:%S',
        # 'filename': LOG_PATH,
        'level': LOG_LEVEL if level is None else level,
    }
    logging.basicConfig(**log_format)
    log = logging.getLogger(APP_ENV if name is None else name)
    log.propagate = True


init_logger(name=APP_ENV, level=logging.DEBUG if DEBUG else None)


def country_blurb(phrase: str) -> str:
    """Return the first word in the hope it is meaningful."""
    if COUNTRY_PAGE:
        return COUNTRY_PAGE.lower()
    return phrase.split()[0].lower()



"""Render the prefix page apps from the prefix abd prefix table stores."""
import collections
import copy
import datetime as dti
import json
import operator
import os
import pathlib
import sys
from typing import List, Union

import mapology.country as cc
import mapology.db as db
import mapology.hull as hull
import mapology.template_loader as template
from mapology import BASE_URL, DEBUG, ENCODING, FOOTER_HTML, FS_PREFIX_PATH, LIB_PATH, PATH_NAV, country_blurb, log

THIS_YY_INT = int(dti.datetime.utcnow().strftime('%y'))

HTML_TEMPLATE = os.getenv('GEO_PREFIX_HTML_TEMPLATE', '')
HTML_TEMPLATE_IS_EXTERNAL = bool(HTML_TEMPLATE)
if not HTML_TEMPLATE:
    HTML_TEMPLATE = 'prefix_page_template.html'

AIRP = 'airport'
RUNW = 'runways'
FREQ = 'frequencies'
LOCA = 'localizers'
GLID = 'glideslopes'

CC_HINT = 'CC_HINT'
City = 'City'
CITY = City.upper()
ICAO = 'ICAO'
IC_PREFIX = 'IC_PREFIX'
IC_PREFIX_ICAO = f'{IC_PREFIX}_{ICAO}'
ITEM = 'ITEM'
KIND = 'KIND'
PATH = '/PATH'
BASE_URL_TARGET = 'BASE_URL'
ANCHOR = 'ANCHOR'
TEXT = 'TEXT'
URL = 'URL'
ZOOM = 'ZOOM'
DEFAULT_ZOOM = 4
FOOTER_HTML_KEY = 'FOOTER_HTML'
LIB_PATH_KEY = 'LIB_PATH'

icao = 'icao_lower'
LAT_LON = 'LAT_LON'
cc_page = 'cc_page'
Cc_page = 'Cc_page'

ATTRIBUTION = f'{KIND} {ITEM} of '

Point = collections.namedtuple('Point', ['label', 'lat', 'lon'])

# GOOGLE_MAPS_URL = f'https://www.google.com/maps/search/?api=1&query={{lat}}%2c{{lon}}'  # Map + pin Documented
GOOGLE_MAPS_URL = 'https://maps.google.com/maps?t=k&q=loc:{lat}+{lon}'  # Sat + pin Undocumented

REGION_COUNTRY_DATA = {
    'region': '',
    'url': '',
    'country': '',
}


def main(argv: Union[List[str], None] = None) -> int:
    """Drive the prefix renderings."""
    argv = sys.argv[1:] if argv is None else argv
    if argv:
        print('usage: mapology prefix')
        return 2

    store_index = db.load_index('store')
    table_index = db.load_index('table')
    hulls_index = db.load_index('hulls')
    region_country_index = db.load_index('region_country')

    prefix_hull_store = copy.deepcopy(hull.THE_HULLS)
    slash = '/'
    prefixes = sorted(store_index.keys())
    num_prefixes = len(prefixes)
    many = num_prefixes > 10  # tenfold magic
    numbers = ('latitude', 'longitude', 'elevation')
    for current, prefix in enumerate(sorted(prefixes), start=1):
        with open(store_index[prefix], 'rt', encoding=ENCODING) as handle:
            prefix_store = json.load(handle)

        with open(table_index[prefix], 'rt', encoding=ENCODING) as handle:
            table_store = json.load(handle)

        hulls_index[prefix] = db.hull_path(prefix)  # noqa

        region_name = table_store['name']
        cc_hint = cc.FROM_ICAO_PREFIX.get(prefix, 'No Country Entry Present')
        my_prefix_path = f'{FS_PREFIX_PATH}/{prefix}'

        region_country_index[prefix] = str(db.DB_FOLDER_PATHS['region_country'] / f'{prefix}.json')
        region_country_data = copy.deepcopy(REGION_COUNTRY_DATA)
        region_country_data['region'] = prefix
        region_country_data['url'] = f'{BASE_URL}/{FS_PREFIX_PATH}/{prefix}/'
        region_country_data['country'] = f'{cc_hint}'
        with open(region_country_index[prefix], 'wt', encoding=ENCODING) as handle:
            json.dump(region_country_data, handle, indent=2)
        log.info(str(region_country_data))

        airports = sorted(table_store['airports'], key=operator.itemgetter('icao'))

        message = f'processing {current :>3d}/{num_prefixes} {prefix} --> ({region_name}) ...'
        if not many or not current % 10 or current == num_prefixes:
            log.info(message)

        if DEBUG:
            log.debug('%s - %s' % (prefix, region_name))
        data_rows = []
        trial_coords = []
        for airport in airports:
            trial_coords.append((airport['latitude'], airport['longitude']))
            row = [str(cell) if key not in numbers else f'{round(cell, 3) :7.03f}' for key, cell in airport.items()]
            # monkey patching
            # ensure cycles are state with two digits zero left padded
            year, cyc = row[6].split(slash)
            row[6] = f'{year}/{int(cyc) :02d}'
            # create a link to the airport page on the ICAO cell of the airport in the table row
            # example: '<a href="AGAT/" class="nd" title="AGAT(Atoifi, Solomon Islands)">AGAT</a>'
            an_icao = row[2]
            a_name = row[8]
            row[2] = f'<a href="{an_icao}/" class="nd" title="{a_name}">{an_icao}</a>'
            if DEBUG:
                log.info('- | %s |' % (' | '.join(row)))
            data_rows.append(
                f'<tr><td>{row[0]}</td><td>{row[1]}</td><td>{row[2]}</td>'
                f'<td class="ra">{row[3]}</td><td class="ra">{row[4]}</td><td class="ra">{row[5]}</td>'
                f'<td class="la">{row[6]}</td><td class="ra">{row[7]}</td>'
                f'<td class="la">{row[8]}</td></tr>'
            )

        prefix_hull = hull.extract_prefix_hull(prefix, region_name, trial_coords)
        hull.update_hull_store(prefix_hull_store, prefix_hull)

        min_lat, min_lon = 90, 180
        max_lat, max_lon = -90, -180
        ra_count = 0  # region_airports_count
        cc_count = 1  # HACK A DID ACK TODO: do not fix country count to wun
        for airport in airports:
            ra_count += 1
            # name has eg. "<a href='KLGA/' target='_blank' title='KLGA(La Guardia, New York, USA)'>KLGA</a>"
            # name_mix = airport['name']
            # name_mix has eg. "KLGA(La Guardia, New York, USA)'>KLGA</a>" LATER TODO
            # code, rest = name_mix.split('(', 1)
            # name = rest.split(')', 1)[0]
            # coords = airport["geometry"]["coordinates"]
            lon, lat = airport['longitude'], airport['latitude']
            min_lat = min(min_lat, lat)
            min_lon = min(min_lon, lon)
            max_lat = max(max_lat, lat)
            max_lon = max(max_lon, lon)

        prefix_lat = 0.5 * (max_lat + min_lat)
        prefix_lon = 0.5 * (max_lon + min_lon)
        bbox_disp = f'[({", ".join(f"{round(v, 3) :7.03f}" for v in (min_lat, min_lon,max_lat, max_lon))}]'
        log.debug('Identified bounding box lat, lon in %s for prefix %s' % (bbox_disp, prefix))
        log.debug(('Set center of prefix map to lat, lon = (%f, %f) for prefix %s' % (prefix_lat, prefix_lon, prefix)))
        prefix_root = pathlib.Path(FS_PREFIX_PATH)
        map_folder = pathlib.Path(prefix_root, prefix)
        map_folder.mkdir(parents=True, exist_ok=True)
        geojson_path = str(pathlib.Path(map_folder, f'{prefix.lower()}-geo.json'))
        with open(geojson_path, 'wt', encoding=ENCODING) as geojson_handle:
            json.dump(prefix_store, geojson_handle, indent=2)

        html_dict = {
            ANCHOR: f'prefix/{prefix}/',
            CC_HINT: cc_hint,
            cc_page: country_blurb(cc_hint),
            Cc_page: country_blurb(cc_hint).title(),
            LAT_LON: f'{prefix_lat},{prefix_lon}',
            LIB_PATH_KEY: LIB_PATH,
            PATH: PATH_NAV,
            BASE_URL_TARGET: BASE_URL,
            ZOOM: str(DEFAULT_ZOOM),
            IC_PREFIX: prefix,
            'IrealCAO': ICAO,
            'ic_prefix_lower-geo.json': f'{prefix.lower()}-geo.json',
            'REGION_AIRPORT_COUNT_DISPLAY': f'{ra_count} airport{"" if ra_count == 1 else "s"}',
            'COUNTRY_COUNT_DISPLAY': f'{cc_count} region{"" if cc_count == 1 else "s"}',
            'BBOX': f' contained in lat, lon bounding box {bbox_disp}',
            FOOTER_HTML_KEY: FOOTER_HTML,
            'DATA_ROWS': '\n'.join(data_rows) + '\n',
        }
        html_page = template.load_html(HTML_TEMPLATE, HTML_TEMPLATE_IS_EXTERNAL)
        for key, replacement in html_dict.items():
            html_page = html_page.replace(key, replacement)

        html_path = pathlib.Path(my_prefix_path, 'index.html')
        with open(html_path, 'wt', encoding=ENCODING) as html_handle:
            html_handle.write(html_page)

        with open(hulls_index[prefix], 'wt', encoding=ENCODING) as handle:
            json.dump(prefix_hull, handle, indent=2)

    db.dump_index('hulls', hulls_index)

    with open(pathlib.Path(FS_PREFIX_PATH) / 'region-hulls-geo.json', 'wt', encoding=ENCODING) as handle:
        json.dump(prefix_hull_store, handle, indent=2)

    db.dump_index('region_country', region_country_index)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



FROM_ICAO_PREFIX = {
    'A1': 'Antarctica (United States)',
    'AG': 'Solomon Islands',
    'AN': 'Nauru',
    'AY': 'Papua New Guinea',
    'BG': 'Greenland',
    'BI': 'Iceland',
    'BK': 'Kosovo',
    'C0': 'Canada',
    'C1': 'Canada',
    'C2': 'Canada',
    'C3': 'Canada',
    'C4': 'Canada',
    'C5': 'Canada',
    'C6': 'Canada',
    'C7': 'Canada',
    'C8': 'Canada',
    'C9': 'Canada',
    'CA': 'Canada',
    'CB': 'Canada',
    'CC': 'Canada',
    'CD': 'Canada',
    'CE': 'Canada',
    'CF': 'Canada',
    'CG': 'Canada',
    'CH': 'Canada',
    'CJ': 'Canada',
    'CK': 'Canada',
    'CL': 'Canada',
    'CM': 'Canada',
    'CN': 'Canada',
    'CO': 'Canada',
    'CP': 'Canada',
    'CR': 'Canada',
    'CS': 'Canada',
    'CT': 'Canada',
    'CV': 'Canada',
    'CW': 'Canada',
    'CX': 'Canada',
    'CY': 'Canada',
    'CZ': 'Canada',
    'DA': 'Algeria',
    'DB': 'Benin',
    'DF': 'Burkina Faso',
    'DG': 'Ghana',
    'DI': "Cote d'Ivoire",
    'DN': 'Nigeria',
    'DR': 'Niger',
    'DT': 'Tunisia',
    'DX': 'Togolese Republic',
    'EB': 'Belgium',
    'ED': 'Germany (civil)',
    'EE': 'Estonia',
    'EF': 'Finland',
    'EG': 'United Kingdom (and Crown dependencies)',
    'EH': 'Netherlands',
    'EI': 'Ireland',
    'EK': 'Denmark and the Faroe Islands',
    'EL': 'Luxembourg',
    'EN': 'Norway',
    'EP': 'Poland',
    'ES': 'Sweden',
    'ET': 'Germany (military)',
    'EV': 'Latvia',
    'EY': 'Lithuania',
    'FA': 'South Africa',
    'FB': 'Botswana',
    'FC': 'Republic of the Congo',
    'FD': 'Eswatini',
    'FE': 'Central African Republic',
    'FG': 'Equatorial Guinea',
    'FH': 'Saint Helena, Ascension and Tristan da Cunha',
    'FI': 'Mauritius',
    'FJ': 'British Indian Ocean Territory',
    'FK': 'Cameroon',
    'FL': 'Zambia',
    'FM': 'Comoros, France (Mayotte and Reunion), and Madagascar',
    'FN': 'Angola',
    'FO': 'Gabon',
    'FP': 'Sao Tome and Príncipe',
    'FQ': 'Mozambique',
    'FS': 'Seychelles',
    'FT': 'Chad',
    'FV': 'Zimbabwe',
    'FW': 'Malawi',
    'FX': 'Lesotho',
    'FY': 'Namibia',
    'FZ': 'Democratic Republic of the Congo',
    'GA': 'Mali',
    'GB': 'The Gambia',
    'GC': 'Spain (Canary Islands)',
    'GE': 'Spain (Ceuta and Melilla)',
    'GF': 'Sierra Leone',
    'GG': 'Guinea-Bissau',
    'GL': 'Liberia',
    'GM': 'Morocco',
    'GO': 'Senegal',
    'GQ': 'Mauritania',
    'GS': 'Western Sahara',
    'GU': 'Guinea',
    'GV': 'Cape Verde',
    'HA': 'Ethiopia',
    'HB': 'Burundi',
    'HC': 'Somalia (including Somaliland)',
    'HD': 'Djibouti',
    'HE': 'Egypt',
    'HH': 'Eritrea',
    'HK': 'Kenya',
    'HL': 'Libya',
    'HR': 'Rwanda',
    'HS': 'Sudan and South Sudan',
    'HT': 'Tanzania',
    'HU': 'Uganda',
    'K0': 'Contiguous United States',
    'K1': 'Contiguous United States',
    'K2': 'Contiguous United States',
    'K3': 'Contiguous United States',
    'K4': 'Contiguous United States',
    'K5': 'Contiguous United States',
    'K6': 'Contiguous United States',
    'K7': 'Contiguous United States',
    'K8': 'Contiguous United States',
    'K9': 'Contiguous United States',
    'KA': 'Contiguous United States',
    'KB': 'Contiguous United States',
    'KC': 'Contiguous United States',
    'KD': 'Contiguous United States',
    'KE': 'Contiguous United States',
    'KF': 'Contiguous United States',
    'KG': 'Contiguous United States',
    'KH': 'Contiguous United States',
    'KI': 'Contiguous United States',
    'KJ': 'Contiguous United States',
    'KK': 'Contiguous United States',
    'KL': 'Contiguous United States',
    'KM': 'Contiguous United States',
    'KN': 'Contiguous United States',
    'KO': 'Contiguous United States',
    'KP': 'Contiguous United States',
    'KR': 'Contiguous United States',
    'KS': 'Contiguous United States',
    'KT': 'Contiguous United States',
    'KU': 'Contiguous United States',
    'KV': 'Contiguous United States',
    'KW': 'Contiguous United States',
    'KX': 'Contiguous United States',
    'KY': 'Contiguous United States',
    'KZ': 'Contiguous United States',
    'LA': 'Albania',
    'LB': 'Bulgaria',
    'LC': 'Cyprus',
    'LD': 'Croatia',
    'LE': 'Spain (mainland section and Balearic Islands)',
    'LF': 'France (Metropolitan France; including Saint-Pierre and Miquelon)',
    'LG': 'Greece',
    'LH': 'Hungary',
    'LI': 'Italy',
    'LJ': 'Slovenia',
    'LK': 'Czech Republic',
    'LL': 'Israel',
    'LM': 'Malta',
    'LN': 'Monaco',
    'LO': 'Austria',
    'LP': 'Portugal (including the Azores and Madeira)',
    'LQ': 'Bosnia and Herzegovina',
    'LR': 'Romania',
    'LS': 'Switzerland',
    'LT': 'Turkey',
    'LU': 'Moldova',
    'LV': 'Palestine/Palestinian territories',
    'LW': 'North Macedonia',
    'LX': 'Gibraltar',
    'LY': 'Serbia and Montenegro',
    'LZ': 'Slovakia',
    'MB': 'Turks and Caicos Islands',
    'MD': 'Dominican Republic',
    'MG': 'Guatemala',
    'MH': 'Honduras',
    'MK': 'Jamaica',
    'MM': 'Mexico',
    'MN': 'Nicaragua',
    'MP': 'Panama',
    'MR': 'Costa Rica',
    'MS': 'El Salvador',
    'MT': 'Haiti',
    'MU': 'Cuba',
    'MW': 'Cayman Islands',
    'MY': 'Bahamas',
    'MZ': 'Belize',
    'NC': 'Cook Islands',
    'NF': 'Fiji, Tonga',
    'NG': 'Kiribati (Gilbert Islands), Tuvalu',
    'NI': 'Niue',
    'NL': 'France (Wallis and Futuna)',
    'NS': 'Samoa, United States (American Samoa)',
    'NT': 'France (French Polynesia)',
    'NV': 'Vanuatu',
    'NW': 'France (New Caledonia)',
    'NZ': 'New Zealand, Antarctica',
    'OA': 'Afghanistan',
    'OB': 'Bahrain',
    'OE': 'Saudi Arabia',
    'OI': 'Iran',
    'OJ': 'Jordan and the West Bank',
    'OK': 'Kuwait',
    'OL': 'Lebanon',
    'OM': 'United Arab Emirates',
    'OO': 'Oman',
    'OP': 'Pakistan',
    'OR': 'Iraq',
    'OS': 'Syria',
    'OT': 'Qatar',
    'OY': 'Yemen',
    'PA': 'US (Alaska)',
    'PB': 'US (Baker Island)',
    'PC': 'Kiribati (Canton Airfield, Phoenix Islands)',
    'PF': 'US (Alaska)',
    'PG': 'US (Guam, Northern Mariana Islands)',
    'PH': 'US (Hawaii)',
    'PJ': 'US (Johnston Atoll)',
    'PK': 'Marshall Islands',
    'PL': 'Kiribati (Line Islands)',
    'PM': 'US (Midway Island)',
    'PO': 'US (Alaska)',
    'PP': 'US (Alaska)',
    'PT': 'Federated States of Micronesia, Palau',
    'PW': 'US (Wake Island)',
    'RC': 'Republic of China (Taiwan)',
    'RJ': 'Japan (Mainland)',
    'RK': 'Republic of Korea (South Korea)',
    'RO': 'Japan (Okinawa)',
    'RP': 'Philippines',
    'S1': 'Argentina',
    'SA': 'Argentina',
    'SB': 'Brazil',
    'SC': 'Chile (including Easter Island)',
    'SD': 'Brazil',
    'SE': 'Ecuador',
    'SF': 'United Kingdom (Falkland Islands)',
    'SG': 'Paraguay',
    'SH': 'Chile',
    'SI': 'Brazil',
    'SJ': 'Brazil',
    'SK': 'Colombia',
    'SL': 'Bolivia',
    'SM': 'Suriname',
    'SN': 'Brazil',
    'SO': 'France (French Guiana)',
    'SP': 'Peru',
    'SS': 'Brazil',
    'SU': 'Uruguay',
    'SV': 'Venezuela',
    'SW': 'Brazil',
    'SY': 'Guyana',
    'TA': 'Antigua and Barbuda',
    'TB': 'Barbados',
    'TD': 'Dominica',
    'TF': 'France (Guadeloupe, Martinique, Saint Barthélemy, Saint Martin)',
    'TG': 'Grenada',
    'TI': 'US (U.S. Virgin Islands)',
    'TJ': 'US (Puerto Rico)',
    'TK': 'Saint Kitts and Nevis',
    'TL': 'Saint Lucia',
    'TN': 'Caribbean Netherlands, Aruba, Curaçao, Sint Maarten',
    'TQ': 'UK (Anguilla)',
    'TR': 'UK (Montserrat)',
    'TT': 'Trinidad and Tobago',
    'TU': 'UK (British Virgin Islands)',
    'TV': 'Saint Vincent and the Grenadines',
    'TX': 'UK (Bermuda)',
    'U0': 'Russia',
    'U1': 'Russia',
    'U3': 'Russia',
    'U4': 'Russia',
    'U5': 'Russia',
    'U6': 'Russia',
    'U7': 'Russia',
    'U8': 'Russia',
    'U9': 'Russia',
    'UA': 'Kazakhstan',
    'UB': 'Azerbaijan',
    'UC': 'Kyrgyzstan',
    'UD': 'Armenia',
    'UE': 'Russia',
    'UF': 'Russia',
    'UG': 'Georgia',
    'UH': 'Russia',
    'UI': 'Russia',
    'UK': 'Ukraine',
    'UL': 'Russia',
    'UM': 'Belarus and Russia (Kaliningrad Oblast)',
    'UN': 'Russia',
    'UO': 'Russia',
    'UR': 'Russia',
    'US': 'Russia',
    'UT': 'Tajikistan, Turkmenistan, Uzbekistan',
    'UU': 'Russia',
    'UW': 'Russia',
    'VA': 'West India',
    'VC': 'Sri Lanka',
    'VD': 'Cambodia',
    'VE': 'East India',
    'VG': 'Bangladesh',
    'VH': 'Hong Kong',
    'VI': 'North India',
    'VL': 'Laos',
    'VM': 'Macau',
    'VN': 'Nepal',
    'VO': 'South India',
    'VQ': 'Bhutan',
    'VR': 'Maldives',
    'VT': 'Thailand',
    'VV': 'Vietnam',
    'VY': 'Myanmar',
    'WA': 'Indonesia',
    'WB': 'Brunei, Malaysia (East Malaysia)',
    'WI': 'Indonesia',
    'WM': 'Malaysia (Peninsular Malaysia)',
    'WP': 'Timor-Leste',
    'WQ': 'Indonesia',
    'WR': 'Indonesia',
    'WS': 'Singapore',
    'Y0': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'Y1': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'Y2': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'Y3': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'Y4': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'Y5': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'Y6': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'Y7': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'Y8': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'Y9': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YA': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YB': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YC': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YD': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YE': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YF': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YG': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YH': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YI': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YJ': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YK': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YL': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YM': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YN': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YO': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YP': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YQ': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YR': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YS': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YT': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YU': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YV': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YW': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'YY': 'Australia (including Norfolk Island, Christmas Island and Cocos (Keeling) Islands)',
    'Z0': 'Mainland China)',
    'Z1': 'Mainland China)',
    'Z2': 'Mainland China)',
    'Z4': 'Mainland China)',
    'Z5': 'Mainland China)',
    'Z8': 'Mainland China)',
    'Z9': 'Mainland China)',
    'ZB': 'Mainland China)',
    'ZG': 'Mainland China)',
    'ZH': 'Mainland China)',
    'ZJ': 'Mainland China)',
    'ZK': 'North Korea',
    'ZL': 'Mainland China)',
    'ZM': 'Mongolia',
    'ZP': 'Mainland China)',
    'ZS': 'Mainland China)',
    'ZU': 'Mainland China)',
    'ZW': 'Mainland China)',
    'ZY': 'Mainland China)',
    'ZZ': 'Unknown Country',
}



"""Determine hull curves from point sets."""
import copy
import functools
from typing import no_type_check

from mapology import log

THE_HULLS = {
    'type': 'FeatureCollection',
    'name': 'Prefix Region Convex Hulls',
    'crs': {
        'type': 'name',
        'properties': {
            'name': 'urn:ogc:def:crs:OGC:1.3:CRS84',
        },
    },
    'features': [],
}
HULL_TEMPLATE = {
    'type': 'Feature',
    'id': '',
    'properties': {'name': ''},
    'geometry': {'type': 'Polygon', 'coordinates': []},
}


@no_type_check
def convex(coords):
    """Executes scan to return points in counter-clockwise order that are on the convex hull (Graham)."""
    turn_left, turn_right, turn_none = 1, -1, 0  # noqa

    def compare(a, b):
        return float(a > b) - float(a < b)

    def turn(p, q, r):
        return compare((q[0] - p[0]) * (r[1] - p[1]) - (r[0] - p[0]) * (q[1] - p[1]), 0)

    def keep_left(hull, r):
        while len(hull) > 1 and turn(hull[-2], hull[-1], r) != turn_left:
            hull.pop()
        if not len(hull) or hull[-1] != r:
            hull.append(r)
        return hull

    points = sorted(coords)
    lower_hull = functools.reduce(keep_left, points, [])
    upper_hull = functools.reduce(keep_left, reversed(points), [])
    return lower_hull.extend(upper_hull[i] for i in range(1, len(upper_hull) - 1)) or lower_hull


@no_type_check
def extract_prefix_hull(prefix, region_name, coords):
    prefix_hull = copy.deepcopy(HULL_TEMPLATE)
    prefix_hull['id'] = prefix
    prefix_hull['properties']['name'] = region_name

    hull_coords = [[lon, lat] for lat, lon in convex(coords)]

    lon_min = min(lon for lon, _ in hull_coords)
    lon_max = max(lon for lon, _ in hull_coords)
    if lon_min < -120 and +120 < lon_max:
        log.info('Problematic region found (convex hull crossing the outer sign change of longitudes): %s' % prefix)
        lon_sign = tuple(+1 if lon > 0 else -1 for lon, _ in hull_coords)
        lon_neg_count = sum(-datum for datum in lon_sign if datum < 0)
        lon_pos_count = sum(datum for datum in lon_sign if datum > 0)
        if lon_pos_count and lon_neg_count:  # Does the polygon cross the longitude sign change?
            log.info('- (%d positive / %d negative) longitudes: %s' % (lon_pos_count, lon_neg_count, prefix))
            patch_me_up = lon_neg_count > lon_pos_count
            add_me = +360 if patch_me_up else -360
            if patch_me_up:
                log.info('- patching upwards (adding 360 degrees to longitude where negative): %s' % prefix)
                hull_coords = [[lon + add_me, lat] if lon < 0 else [lon, lat] for lon, lat in hull_coords]
            else:
                log.info('- patching downwards (adding 360 degrees to longitude where positive): %s' % prefix)
                hull_coords = [[lon + add_me, lat] if lon > 0 else [lon, lat] for lon, lat in hull_coords]

    if prefix == 'ET':
        log.info('Patching a North-Eastern ear onto the hull: %s' % prefix)

        @no_type_check
        def et_earify(pair):
            """Monkey patch the ET region to offer an ear to select outside of ED."""
            lon_ne = 12.27  # 9333333333334,
            lat_ne = 53.91  # 8166666666664
            magic_lon = 13.648875
            magic_lat = 54.551359
            lon, lat = pair
            return [magic_lon, magic_lat] if lon > lon_ne and lat > lat_ne else [lon, lat]

        hull_coords = [et_earify(pair) for pair in hull_coords]

    # problem regions A1, NZ, NF, PA, UH,
    prefix_hull['geometry']['coordinates'].append(hull_coords)
    return prefix_hull


@no_type_check
def update_hull_store(hull_store, prefix_hull):
    """DRY."""
    hull_store['features'].append(copy.deepcopy(prefix_hull))
